# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import math
from typing import List, Optional, Sequence, TYPE_CHECKING

import numpy
from numpy import integer, uint32
from numpy.typing import NDArray

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.utilities import utility_calls
from spynnaker.pyNN.exceptions import SpynnakerException

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


class FixedNumberPreConnector(AbstractGenerateConnectorOnMachine,
                              AbstractGenerateConnectorOnHost):
    """
    Connects a fixed number of pre-synaptic neurons selected at random,
    to all post-synaptic neurons.
    """

    __slots__ = (
        "__allow_self_connections",
        "__n_pre",
        "__pre_neurons",
        "__pre_neurons_set",
        "__with_replacement",
        "__rng")

    def __init__(
            self, n: int, *, allow_self_connections: bool = True,
            with_replacement: bool = False, rng: Optional[NumpyRNG] = None,
            safe: bool = True, verbose: bool = False, callback: None = None):
        """
        :param n:
            number of random pre-synaptic neurons connected to output
        :param allow_self_connections:
            if the connector is used to connect a Population to itself,
            this flag determines whether a neuron is allowed to connect to
            itself, or only to other neurons in the Population.
        :param safe:
            Whether to check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param with_replacement:
            this flag determines how the random selection of pre-synaptic
            neurons is performed; if true, then every pre-synaptic neuron
            can be chosen on each occasion, and so multiple connections
            between neuron pairs are possible; if false, then once a
            pre-synaptic neuron has been connected to a post-neuron, it
            can't be connected again.
        :param rng:
            Seeded random number generator, or `None` to make one when needed
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)
        # We absolutely require an integer at this point!
        self.__n_pre = self._roundsize(n, "FixedNumberPreConnector")
        self.__allow_self_connections = allow_self_connections
        self.__with_replacement = with_replacement
        self.__pre_neurons_set = False
        self.__pre_neurons: List[NDArray[integer]] = []
        self.__rng = rng

    def set_projection_information(
            self, synapse_info: SynapseInformation) -> None:
        super().set_projection_information(synapse_info)
        if (not self.__with_replacement and
                self.__n_pre > synapse_info.n_pre_neurons):
            raise SpynnakerException(
                "FixedNumberPreConnector will not work when "
                "with_replacement=False and n > n_pre_neurons")

        if (not self.__with_replacement and
                not self.__allow_self_connections and
                self.__n_pre == synapse_info.n_pre_neurons and
                synapse_info.pre_population is synapse_info.post_population):
            raise SpynnakerException(
                "FixedNumberPreConnector will not work when "
                "with_replacement=False, allow_self_connections=False "
                "and n = n_pre_neurons")

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, self.__n_pre * synapse_info.n_post_neurons,
            synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, self.__n_pre * synapse_info.n_post_neurons,
            synapse_info)

    def __build_pre_neurons(self, synapse_info: SynapseInformation) -> List[
            NDArray[integer]]:
        rng = self.__rng or NumpyRNG()
        pre_neurons: List[NDArray[integer]] = [
            numpy.zeros([0], dtype=uint32)] * synapse_info.n_post_neurons
        # Loop over all the post neurons
        for m in range(synapse_info.n_post_neurons):
            # If the pre and post populations are the same
            # then deal with allow_self_connections=False
            if (synapse_info.pre_population is synapse_info.post_population
                    and not self.__allow_self_connections):
                # Exclude the current pre-neuron from the post-neuron
                # list
                no_self_pre_neurons = [
                    n for n in range(synapse_info.n_pre_neurons) if n != m]

                # Now use this list in the random choice
                pre_neurons[m] = rng.choice(
                    no_self_pre_neurons, self.__n_pre,
                    self.__with_replacement)
            else:
                pre_neurons[m] = rng.choice(
                    synapse_info.n_pre_neurons, self.__n_pre,
                    self.__with_replacement)

            # Sort the neurons now that we have them
            pre_neurons[m].sort()
        return pre_neurons

    def _get_pre_neurons(self, synapse_info: SynapseInformation) -> List[
            NDArray[integer]]:
        # If we haven't set the array up yet, do it now
        if not self.__pre_neurons_set:
            self.__pre_neurons = self.__build_pre_neurons(synapse_info)
            self.__pre_neurons_set = True

            # if verbose open a file to output the connectivity
            if self.verbose:
                filename = f"{synapse_info.pre_population.label}_to_" \
                           f"{synapse_info.post_population.label}" \
                           f"_fixednumberpre-conn.csv"
                with open(filename, 'w', encoding="utf-8") as file_handle:
                    numpy.savetxt(file_handle,  # type: ignore[arg-type]
                                  [(synapse_info.n_pre_neurons,
                                    synapse_info.n_post_neurons,
                                    self.__n_pre)],
                                  fmt="%u,%u,%u")
                    for pre_neuron in self.__pre_neurons:
                        numpy.savetxt(file_handle,  # type: ignore[arg-type]
                                      pre_neuron[None, :],
                                      fmt=("%u," * (self.__n_pre - 1) + "%u"))

        return self.__pre_neurons

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        prob_selection = 1.0 / float(synapse_info.n_pre_neurons)
        n_connections_total = utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            self.__n_pre * synapse_info.n_post_neurons, prob_selection,
            chance=1.0/10000.0)
        prob_in_slice = min(
            float(n_post_atoms) / float(synapse_info.n_post_neurons), 1.0)
        n_connections = utility_calls.get_probable_maximum_selected(
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections_total, prob_in_slice, chance=1.0/100000.0)

        if min_delay is None or max_delay is None:
            return int(math.ceil(n_connections))

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        return self.__n_pre

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
           synapse_info.weights, self.__n_pre * synapse_info.n_post_neurons,
           synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        # Get lo and hi for the post vertex
        lo = post_vertex_slice.lo_atom
        hi = post_vertex_slice.hi_atom

        pre_neurons = self._get_pre_neurons(synapse_info)

        # Get number of connections
        n_connections = self.__n_pre * post_vertex_slice.n_atoms

        # Set up the block
        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)

        # Set up source and target
        pre_neurons_in_slice: List[NDArray[integer]] = []
        post_neurons_in_slice: List[NDArray[integer]] = []
        post_vertex_array = numpy.arange(lo, hi + 1)
        for n in range(lo, hi + 1):
            for pn in pre_neurons[n]:
                pre_neurons_in_slice.append(pn)
                post_neurons_in_slice.append(post_vertex_array[n - lo])

        block["source"] = pre_neurons_in_slice
        block["target"] = post_neurons_in_slice

        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"FixedNumberPreConnector({self.__n_pre})"

    @property
    def allow_self_connections(self) -> bool:
        """
        Do we include connections from a neuron/id to itself?
        """
        return self.__allow_self_connections

    @allow_self_connections.setter
    def allow_self_connections(self, new_value: bool) -> None:
        self.__allow_self_connections = new_value

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.FIXED_NUMBER_PRE_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        allow_self = (
            self.__allow_self_connections or
            synapse_info.pre_population != synapse_info.post_population)
        return numpy.array([
            int(allow_self),
            int(self.__with_replacement),
            self.__n_pre], dtype=uint32)

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return 3 * BYTES_PER_WORD

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if self.generate_on_machine(synapse_info):
            utility_calls.check_rng(self.__rng, "FixedNumberPreConnector")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections import Counter
import logging
from typing import (Any, Collection, Dict, List, Optional, Sequence, Tuple,
                    Union, TYPE_CHECKING)

import numpy
from numpy.typing import ArrayLike, NDArray
from typing_extensions import TypeAlias, TypeGuard

from pyNN.space import Grid2D, Grid3D, BaseStructure

from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.config_holder import get_config_int
from spinn_utilities.ranged.abstract_sized import Selector

from pacman.model.graphs.common import Slice
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.model.resources import AbstractSDRAM

from spinn_front_end_common.utility_models import ReverseIpTagMultiCastSource

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.abstract_models import SupportsStructure
from spynnaker.pyNN.models.common import (
    ParameterHolder, PopulationApplicationVertex)
from spynnaker.pyNN.models.common.types import (Names, Spikes)
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType
from spynnaker.pyNN.utilities.ranged import SpynnakerRangedList

from .spike_source_array_machine_vertex import SpikeSourceArrayMachineVertex

if TYPE_CHECKING:
    from .spike_source_array import SpikeSourceArray

logger = FormatAdapter(logging.getLogger(__name__))

# Cut off to warn too many spikes sent at one time
TOO_MANY_SPIKES = 100

_Number: TypeAlias = Union[int, float]

_SingleList: TypeAlias = Union[
    Sequence[_Number], NDArray[numpy.integer]]
_DoubleList: TypeAlias = Union[
    Sequence[Sequence[_Number]], NDArray[numpy.integer]]


def _is_double_list(value: Spikes) -> TypeGuard[_DoubleList]:
    return not isinstance(value, (float, int)) and bool(len(value)) and \
        hasattr(value[0], "__len__")


def _is_single_list(value: Spikes) -> TypeGuard[_SingleList]:
    # USE _is_double_list first!
    return not isinstance(value, (float, int)) and bool(len(value))


def _is_singleton(value: Spikes) -> TypeGuard[_Number]:
    return isinstance(value, (float, int))


def _as_numpy_ticks(
        times: ArrayLike, time_step: float) -> NDArray[numpy.int64]:
    return numpy.ceil(
        numpy.floor(numpy.array(times) * 1000.0) / time_step).astype("int64")


def _send_buffer_times(
        spike_times: Spikes, time_step: float) -> Union[
            NDArray[numpy.int64], List[NDArray[numpy.int64]]]:
    # Convert to ticks
    if _is_double_list(spike_times):
        return [_as_numpy_ticks(times, time_step) for times in spike_times]
    elif _is_single_list(spike_times):
        return _as_numpy_ticks(spike_times, time_step)
    elif _is_singleton(spike_times):
        return _as_numpy_ticks([spike_times], time_step)
    else:
        return []


class SpikeSourceArrayVertex(
        ReverseIpTagMultiCastSource, PopulationApplicationVertex,
        SupportsStructure):
    """
    Model for play back of spikes.
    """
    __slots__ = (
        "__model_name",
        "__model",
        "__structure",
        "_spike_times",
        "__n_colour_bits")

    #: ID of the recording region used for recording transmitted spikes.
    SPIKE_RECORDING_REGION_ID = 0

    def __init__(
            self, n_neurons: int, spike_times: Spikes, label: str,
            max_atoms_per_core: Union[int, Tuple[int, ...]],
            model: SpikeSourceArray,
            splitter: Optional[AbstractSplitterCommon],
            n_colour_bits: Optional[int]):
        """

        :param n_neurons: The number of neurons in the population
        :param spike_times: Timesteps on which to spike
        :param label:
        :param max_atoms_per_core:
        :param model:
        :param splitter:
        :param n_colour_bits:
        """
        self.__model_name = "SpikeSourceArray"
        self.__model = model
        self.__structure: Optional[BaseStructure] = None

        if spike_times is None:
            spike_times = []
        self._spike_times = SpynnakerRangedList(
            n_neurons, spike_times,
            use_list_as_value=not _is_double_list(spike_times))

        time_step = SpynnakerDataView.get_simulation_time_step_us()

        super().__init__(
            n_keys=n_neurons, label=label,
            max_atoms_per_core=max_atoms_per_core,
            send_buffer_times=_send_buffer_times(spike_times, time_step),
            splitter=splitter)

        self._check_spike_density(spike_times)
        # Do colouring
        if n_colour_bits is None:
            self.__n_colour_bits = get_config_int(
                "Simulation", "n_colour_bits")
        else:
            self.__n_colour_bits = n_colour_bits

    @overrides(ReverseIpTagMultiCastSource.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice: Slice, sdram: AbstractSDRAM,
            label: Optional[str] = None) -> SpikeSourceArrayMachineVertex:
        send_buffer_times = self._filtered_send_buffer_times(vertex_slice)
        machine_vertex = SpikeSourceArrayMachineVertex(
            label=label, app_vertex=self, vertex_slice=vertex_slice,
            eieio_params=self._eieio_params,
            send_buffer_times=send_buffer_times)
        machine_vertex.enable_recording(self._is_recording)
        # Known issue with ReverseIPTagMulticastSourceMachineVertex
        if sdram:
            assert sdram == machine_vertex.sdram_required
        return machine_vertex

    def _check_spike_density(self, spike_times: Spikes) -> None:
        if _is_double_list(spike_times):
            self._check_density_double_list(spike_times)
        elif _is_single_list(spike_times):
            self._check_density_single_list(spike_times)
        elif _is_singleton(spike_times):
            pass
        else:
            logger.warning("SpikeSourceArray has no spike times")

    def _check_density_single_list(self, spike_times: _SingleList) -> None:
        counter = Counter(spike_times)
        top = counter.most_common(1)
        val, count = top[0]
        if count * self.n_atoms > TOO_MANY_SPIKES:
            if self.n_atoms > 1:
                logger.warning(
                    "Danger of SpikeSourceArray sending too many spikes "
                    "at the same time. "
                    "This is because ({}) neurons share the same spike list",
                    self.n_atoms)
            else:
                logger.warning(
                    "Danger of SpikeSourceArray sending too many spikes "
                    "at the same time. "
                    "For example at time {}, {} spikes will be sent",
                    val, count * self.n_atoms)

    def _check_density_double_list(self, spike_times: _DoubleList) -> None:
        counter: Counter = Counter()
        for neuron_id in range(0, self.n_atoms):
            counter.update(spike_times[neuron_id])
        top = counter.most_common(1)
        val, count = top[0]
        if count > TOO_MANY_SPIKES:
            logger.warning(
                "Danger of SpikeSourceArray sending too many spikes "
                "at the same time. "
                "For example at time {}, {} spikes will be sent",
                val, count)

    @overrides(SupportsStructure.set_structure)
    def set_structure(self, structure: BaseStructure) -> None:
        self.__structure = structure

    @property
    @overrides(ReverseIpTagMultiCastSource.atoms_shape)
    def atoms_shape(self) -> Tuple[int, ...]:
        if isinstance(self.__structure, (Grid2D, Grid3D)):
            return self.__structure.calculate_size(self.n_atoms)
        return super().atoms_shape

    def _to_early_spikes_single_list(self, spike_times: _SingleList) -> None:
        """
        Checks if there is one or more spike_times before the current time.

        Logs a warning for the first one found

        :param spike_times:
        """
        current_time = SpynnakerDataView.get_current_run_time_ms()
        for spike_time in spike_times:
            if spike_time < current_time:
                logger.warning(
                    "SpikeSourceArray {} has spike_times that are lower than "
                    "the current time {} For example {} - "
                    "these will be ignored.",
                    self, current_time, float(spike_time))
                return

    def _check_spikes_double_list(self, spike_times: _DoubleList) -> None:
        """
        Checks if there is one or more spike_times before the current time.

        Logs a warning for the first one found

        :param spike_times:
        """
        current_time = SpynnakerDataView.get_current_run_time_ms()
        for neuron_id in range(0, self.n_atoms):
            id_times = spike_times[neuron_id]
            for id_time in id_times:
                if id_time < current_time:
                    logger.warning(
                        "SpikeSourceArray {} has spike_times that are lower "
                        "than the current time {} For example {} - "
                        "these will be ignored.",
                        self, current_time, float(id_time))
                    return

    def __set_spike_buffer_times(self, spike_times: Spikes) -> None:
        """
        Set the spike source array's buffer spike times.
        """
        time_step = SpynnakerDataView.get_simulation_time_step_us()
        # warn the user if they are asking for a spike time out of range
        if _is_double_list(spike_times):
            self._check_spikes_double_list(spike_times)
        elif _is_single_list(spike_times):
            self._to_early_spikes_single_list(spike_times)
        elif _is_singleton(spike_times):
            self._to_early_spikes_single_list([spike_times])
        else:
            # in case of empty list do not check
            pass
        self.send_buffer_times = _send_buffer_times(spike_times, time_step)
        self._check_spike_density(spike_times)

    def __read_parameter(self, name: str, selector: Selector) -> Sequence:
        _ = name
        return self._spike_times.get_values(selector)

    @overrides(PopulationApplicationVertex.get_parameter_values)
    def get_parameter_values(
            self, names: Names, selector: Selector = None) -> ParameterHolder:
        self._check_parameters(names, {"spike_times"})
        return ParameterHolder(names, self.__read_parameter, selector)

    @overrides(PopulationApplicationVertex.set_parameter_values)
    def set_parameter_values(
            self, name: str, value: Spikes, selector: Selector = None) -> None:
        self._check_parameters(name, {"spike_times"})
        self.__set_spike_buffer_times(value)
        self._spike_times.set_value_by_selector(
            selector, value, use_list_as_value=not _is_double_list(value))

    @overrides(PopulationApplicationVertex.get_parameters)
    def get_parameters(self) -> List[str]:
        return ["spike_times"]

    @overrides(PopulationApplicationVertex.get_units)
    def get_units(self, name: str) -> str:
        if name == "spikes":
            return ""
        if name == "spike_times":
            return "ms"
        raise KeyError(f"Units for {name} unknown")

    @overrides(PopulationApplicationVertex.get_recordable_variables)
    def get_recordable_variables(self) -> List[str]:
        return ["spikes"]

    @overrides(PopulationApplicationVertex.get_buffer_data_type)
    def get_buffer_data_type(self, name: str) -> BufferDataType:
        if name == "spikes":
            return BufferDataType.EIEIO_SPIKES
        raise KeyError(f"Cannot record {name}")

    @overrides(PopulationApplicationVertex.get_neurons_recording)
    def get_neurons_recording(
            self, name: str, vertex_slice: Slice) -> NDArray[numpy.integer]:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return vertex_slice.get_raster_ids()

    @overrides(PopulationApplicationVertex.set_recording)
    def set_recording(
            self, name: str, sampling_interval: Optional[float] = None,
            indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if sampling_interval is not None:
            logger.warning("Sampling interval currently not supported for "
                           "SpikeSourceArray so being ignored")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeSourceArray so being ignored")
        self.enable_recording(True)
        SpynnakerDataView.set_requires_mapping()

    @overrides(PopulationApplicationVertex.set_not_recording)
    def set_not_recording(self, name: str,
                          indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeSourceArray so being ignored")
        self.enable_recording(False)

    @overrides(PopulationApplicationVertex.get_recording_variables)
    def get_recording_variables(self) -> List[str]:
        if self._is_recording:
            return ["spikes"]
        return []

    @overrides(PopulationApplicationVertex.get_sampling_interval_ms)
    def get_sampling_interval_ms(self, name: str) -> float:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return SpynnakerDataView.get_simulation_time_step_us()

    @overrides(PopulationApplicationVertex.get_recording_region)
    def get_recording_region(self, name: str) -> int:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return self.SPIKE_RECORDING_REGION_ID

    @overrides(PopulationApplicationVertex.get_data_type)
    def get_data_type(self, name: str) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return None

    def describe(
            self) -> Dict[str, Union[str, ParameterHolder, Dict[str, Any]]]:
        """
        Returns a human-readable description of the cell or synapse type.

        The output may be customised by specifying a different template
        together with an associated template engine
        (see :py:mod:`pyNN.descriptions`).

        If template is `None`, then a dictionary containing the template
        context will be returned.

        :returns: A human-readable description of the cell
        """
        return {
            "name": self.__model_name,
            "default_parameters": self.__model.default_parameters,
            "default_initial_values": self.__model.default_parameters,
            "parameters": self.get_parameter_values(
                self.__model.default_parameters),
        }

    @property
    @overrides(PopulationApplicationVertex.n_colour_bits)
    def n_colour_bits(self) -> int:
        return self.__n_colour_bits



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Decorators to support default argument handling.
"""

import inspect
import logging
from types import MappingProxyType
from typing import (
    Any, Callable, FrozenSet, Iterable, List, Mapping, Optional, Tuple)
from spinn_utilities.classproperty import classproperty
from spinn_utilities.log import FormatAdapter

logger = FormatAdapter(logging.getLogger(__name__))


def _check_args(
        args_to_find: FrozenSet[str], default_args: List[str],
        init_method: Callable) -> None:
    for arg in args_to_find:
        if arg not in default_args:
            raise AttributeError(
                f"Argument {arg} not found, or "
                f"no default value provided in {init_method}")


def get_map_from_init(
        init_method: Callable,
        skip: Optional[FrozenSet[str]] = None,
        include: Optional[FrozenSet[str]] = None) -> Mapping[str, Any]:
    """
    Get an argument initialisation dictionary by examining an
    ``__init__`` method or function.

    :param init_method: The method.
    :param skip: The arguments to be skipped, if any
    :param include: The arguments that must be present, if any
    :return: an initialisation dictionary
    """
    init_args = inspect.getfullargspec(init_method)
    n_defaults = 0 if init_args.defaults is None else len(init_args.defaults)
    n_args = 0 if init_args.args is None else len(init_args.args)
    default_args = ([] if init_args.args is None else
                    init_args.args[n_args - n_defaults:])
    default_values = [] if init_args.defaults is None else init_args.defaults

    # Check that included / skipped things exist
    if include is not None:
        _check_args(include, default_args, init_method)
    if skip is not None:
        _check_args(skip, default_args, init_method)

    as_dict = {arg: value
               for arg, value in zip(default_args, default_values)
               if ((arg != "self") and
                   (skip is None or arg not in skip) and
                   (include is None or arg in include))}
    return MappingProxyType(as_dict)


def default_parameters(parameters: Iterable[str]) -> Callable:
    """
    Specifies arguments which are parameters.  Only works on the
    ``__init__`` method of a class that is additionally decorated with
    :py:meth:`defaults`

    :param parameters:
        The names of the arguments that are parameters
    :returns: A check method to be called when first used
    """
    def wrap(method: Callable) -> Callable:
        # pylint: disable=protected-access
        # Find the real method in case we use multiple of these decorators
        wrapped = method
        while hasattr(method, "_method"):
            method = getattr(method, "_method")

        # Set the parameters of the method to be used later
        method._parameters = (  # type: ignore[attr-defined]
            frozenset(parameters))
        method_args = inspect.getfullargspec(method)

        def wrapper(*args: Any, **kwargs: Any) -> None:
            # Check for state variables that have been specified in cell_params
            args_provided = method_args.args[:len(args)]
            args_provided.extend([
                arg for arg in kwargs if arg in method_args.args])
            parameters = method._parameters  # type: ignore[attr-defined]
            for arg in args_provided:
                if arg not in parameters and arg != "self":
                    logger.warning(
                        "Formal PyNN specifies that {} should be set using "
                        "initial_values not cell_params", arg)
            wrapped(*args, **kwargs)

        # Store the real method in the returned object
        wrapper._method = method  # type: ignore[attr-defined]
        return wrapper
    return wrap


def default_initial_values(state_variables: Iterable[str]) -> Callable:
    """
    Specifies arguments which are state variables.  Only works on the
    ``__init__`` method of a class that is additionally decorated with
    :py:meth:`defaults`

    :param state_variables:
        The names of the arguments that are state variables
    :returns: A check method to be called when first used
    """
    def wrap(method: Callable) -> Callable:
        """
        Wraps the init method with a check method

        :param method: init method to wrap
        :returns: A check method to be called when first used
        """
        # pylint: disable=protected-access
        # Find the real method in case we use multiple of these decorators
        wrapped = method
        while hasattr(method, "_method"):
            method = getattr(method, "_method")

        # Store the state variables of the method to be used later
        method._state_variables = (  # type: ignore[attr-defined]
            frozenset(state_variables))
        method_args = inspect.getfullargspec(method)

        def wrapper(*args: Any, **kwargs: Any) -> None:
            # Check for state variables that have been specified in cell_params
            args_provided = method_args.args[:len(args)]
            args_provided.extend([
                arg for arg in kwargs if arg in method_args.args])
            variables = method._state_variables  # type: ignore[attr-defined]
            for arg in args_provided:
                if arg in variables:
                    logger.warning(
                        "Formal PyNN specifies that {} should be set using "
                        "initial_values not cell_params", arg)
            wrapped(*args, **kwargs)

        # Store the real method in the returned object
        wrapper._method = method  # type: ignore[attr-defined]
        return wrapper
    return wrap


def defaults(cls: type) -> type:
    """
    Deprecated! Extend AbstractProvidesDefaults instead

    Get the default parameters and state variables from the arguments to
    the ``__init__`` method.  This uses the decorators
    :py:func:`default_parameters` and :py:func:`default_initial_values` to
    determine the parameters and state variables respectively.
    If only one is specified, the other is assumed to be the remaining
    arguments.
    If neither are specified, it is assumed that all default arguments are
    parameters.

    :returns: input unchanged
    """
    logger.warning("@defaults is deprecated! "
                   "Extend AbstractProvidesDefaults instead")
    if not inspect.isclass(cls):
        raise TypeError(f"{cls} is not a class")
    if not hasattr(cls, "__init__"):
        raise AttributeError(f"No __init__ found in {cls}")
    init = getattr(cls, "__init__")
    while hasattr(init, "_method"):
        init = getattr(init, "_method")
    params = None
    if hasattr(init, "_parameters"):
        params = getattr(init, "_parameters")
    svars = None
    if hasattr(init, "_state_variables"):
        svars = getattr(init, "_state_variables")
    if params is None and svars is None:
        cls.default_parameters = get_map_from_init(init)
        cls.default_initial_values = {}
    elif params is None:
        cls.default_parameters = get_map_from_init(init, skip=svars)
        cls.default_initial_values = get_map_from_init(init, include=svars)
    elif svars is None:
        cls.default_parameters = get_map_from_init(init, include=params)
        cls.default_initial_values = get_map_from_init(init, skip=params)
    else:
        cls.default_parameters = get_map_from_init(init, include=params)
        cls.default_initial_values = get_map_from_init(init, include=svars)
    return cls


class AbstractProvidesDefaults(object):
    """
    Provides the default_parameters and default_initial_values properties

    These will be filled in based on the @default_parameters and
    @default_initial_values decorators with values read from the init.
    """

    __cashed_defaults: Optional[Mapping[str, Any]] = None
    __cashed_initials: Optional[Mapping[str, Any]] = None

    @classmethod
    def __fill_in_defaults(cls) -> None:
        """
        Fills in default_parameters and default_initial_values attributes
        """
        # get the init method
        init = getattr(cls, "__init__")
        # Find the real method as there may be decorators
        while hasattr(init, "_method"):
            init = getattr(init, "_method")

        # read the values from the init method
        init_args = inspect.getfullargspec(init)
        n_defaults = 0 if init_args.defaults is None else len(
            init_args.defaults)
        n_args = 0 if init_args.args is None else len(init_args.args)
        default_args = ([] if init_args.args is None else
                        init_args.args[n_args - n_defaults:])
        if init_args.defaults is None:
            default_values: Tuple = ()
        else:
            default_values = init_args.defaults

        # get the keys based on the decorators
        if hasattr(init, "_parameters"):
            params = getattr(init, "_parameters")
            if hasattr(init, "_state_variables"):
                svars = getattr(init, "_state_variables")
                assert len(params.intersection(svars)) == 0
            else:
                svars = frozenset(svar for svar in default_args
                                  if svar not in params)
        else:
            if hasattr(init, "_state_variables"):
                svars = getattr(init, "_state_variables")
                params = frozenset(param for param in default_args
                                   if param not in svars)
            else:
                svars = frozenset()
                params = set(default_args)

        # Check all decorator values used
        _check_args(params.union(svars), default_args, init)

        # fill in the defaults so this method is only called once
        __defaults = {}
        __initials = {}
        for arg, value in zip(default_args, default_values):
            if arg in params:
                __defaults[arg] = value
            elif arg in svars:
                __initials[arg] = value
        cls.__cashed_defaults = MappingProxyType(__defaults)
        cls.__cashed_initials = MappingProxyType(__initials)

    @classproperty
    def default_parameters(  # pylint: disable=no-self-argument
            cls) -> Mapping[str, Any]:
        """
        Get the default values for the parameters of the model.

        If a @default_parameters decorator is used
        this will be the init default values for those keys

        If no @default_parameters decorator is used
        this will be all the init parameters with a default value
        less any defined in @default_initial_values

        :returns: Mapping of parameter names to default values
        """
        if cls.__cashed_defaults is None:
            cls.__fill_in_defaults()
            assert cls.__cashed_defaults is not None
        return cls.__cashed_defaults

    @classproperty
    def default_initial_values(  # pylint: disable=no-self-argument
            cls) -> Mapping[str, Any]:
        """
        Get the default initial values for the state variables of the model.

        If @default_initial_values decorator is used
        this will be the init default values for those keys

        If no @default_initial_values decorator is used
        but a @default_parameters decorator was used
        this will be all the init parameters with a default value
        less any defined in @default_parameters

        If neither decorator is used this will be an empty Mapping

        :returns:
            The default initial values for the state variables of the model.

        """
        if cls.__cashed_initials is None:
            cls.__fill_in_defaults()
            assert cls.__cashed_initials is not None
        return cls.__cashed_initials



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from math import ceil
from typing import (
    Dict, List, Iterable, Tuple, cast, TYPE_CHECKING)

import numpy
from numpy import floating, uint32
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.application import ApplicationVertex

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationGenerator)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.abstract_models import ColouredApplicationVertex
from spynnaker.pyNN.models.neural_projections.connectors import (
    PoolDenseConnector)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSupportsSignedWeights)
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN
from spynnaker.pyNN.models.common.local_only_2d_common import (
    get_sources_for_target, get_rinfo_for_spike_source, BITS_PER_SHORT,
    get_div_const, N_COLOUR_BITS_BITS, KEY_INFO_SIZE, get_first_and_last_slice,
    Source)

from .abstract_local_only import AbstractLocalOnly

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neuron import (
        PopulationMachineLocalOnlyCombinedVertex)
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)

#: Size of the source information
SOURCE_INFO_SIZE = KEY_INFO_SIZE + BYTES_PER_WORD

#: Size of the source info per-dimension info
SOURCE_INFO_DIM_SIZE = 9 * BYTES_PER_WORD

#: Size of information
CONFIG_SIZE = 3 * BYTES_PER_WORD


class LocalOnlyPoolDense(AbstractLocalOnly, AbstractSupportsSignedWeights):
    """
    A convolution synapse dynamics that can process spikes with only DTCM.
    """

    __slots__ = [
        "__cached_sources"]

    def __init__(self, delay: WEIGHTS_DELAYS_IN = None):
        """
        :param delay:
            The delay used in the connection; by default 1 time step
        """
        # Store the sources to avoid recalculation
        self.__cached_sources: Dict[ApplicationVertex, Dict[
                Tuple[ColouredApplicationVertex, str], List[Source]]] = dict()

        super().__init__(delay)
        if not isinstance(self.delay, (float, int)):
            raise SynapticConfigurationException(
                "Only single value delays are supported")

    @property
    def _delay(self) -> float:
        # Guaranteed by check in init
        return cast(float, self.delay)

    @overrides(AbstractLocalOnly.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> LocalOnlyPoolDense:
        if not isinstance(synapse_dynamics, LocalOnlyPoolDense):
            raise SynapticConfigurationException(
                "All Projections of this Population must have a synapse_type"
                " of LocalOnlyPoolDense")
        return synapse_dynamics

    @overrides(AbstractLocalOnly.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return "_pool_dense"

    @property
    @overrides(AbstractLocalOnly.changes_during_run)
    def changes_during_run(self) -> bool:
        return False

    @staticmethod
    def __connector(projection: Projection) -> PoolDenseConnector:
        # pylint: disable=protected-access
        return cast(PoolDenseConnector,
                    projection._synapse_information.connector)

    @overrides(AbstractLocalOnly.get_parameters_usage_in_bytes)
    def get_parameters_usage_in_bytes(
            self, n_atoms: int,
            incoming_projections: Iterable[Projection]) -> int:
        n_bytes = 0
        seen_edges = set()
        for incoming in incoming_projections:
            # pylint: disable=protected-access
            s_info = incoming._synapse_information
            if not isinstance(s_info.connector, PoolDenseConnector):
                raise SynapticConfigurationException(
                    "Only PoolDenseConnector can be used with a synapse type"
                    " of PoolDense")
            # pylint: disable=protected-access
            app_edge = incoming._projection_edge
            if app_edge not in seen_edges:
                seen_edges.add(app_edge)
                n_dims = len(app_edge.pre_vertex.atoms_shape)
                n_bytes += SOURCE_INFO_SIZE
                n_bytes += n_dims * SOURCE_INFO_DIM_SIZE
            n_bytes += s_info.connector.local_only_n_bytes(
                app_edge.pre_vertex.atoms_shape, n_atoms)

        return CONFIG_SIZE + n_bytes

    @overrides(AbstractLocalOnly.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationGenerator, region: int,
            machine_vertex: PopulationMachineLocalOnlyCombinedVertex,
            weight_scales: NDArray[floating]) -> None:
        # Get incoming sources for this vertex
        app_vertex = cast('PopulationVertex',
                          machine_vertex.app_vertex)
        sources = self.__get_sources_for_target(app_vertex)

        size = self.get_parameters_usage_in_bytes(
            machine_vertex.vertex_slice.n_atoms,
            app_vertex.incoming_projections)
        spec.reserve_memory_region(region, size, label="LocalOnlyPoolDense")
        spec.switch_write_focus(region)

        connector_data: List[NDArray[uint32]] = list()
        source_data = list()
        n_connectors = 0
        for (pre_vertex, part_id), source_infos in sources.items():
            first_conn_index = len(connector_data)
            for source in source_infos:
                # pylint: disable=protected-access
                conn = cast(PoolDenseConnector,
                            source.projection._synapse_information.connector)
                app_edge = source.projection._projection_edge
                connector_data.append(conn.get_local_only_data(
                    app_edge, source.local_delay, source.delay_stage,
                    machine_vertex.vertex_slice, weight_scales))
                n_connectors += 1

            # Get the source routing information
            r_info, core_mask, mask_shift = get_rinfo_for_spike_source(
                pre_vertex, part_id)

            # Get the width / height per core / last_core
            first_slice, last_slice = get_first_and_last_slice(pre_vertex)
            n_dims = len(pre_vertex.atoms_shape)
            pre_shape = list(pre_vertex.atoms_shape)

            # Add the key and mask...
            source_data.extend([r_info.key, r_info.mask])
            # ... start connector index, n_colour_bits, count of connectors ...
            source_data.append(
                (len(source_infos) << BITS_PER_SHORT) +
                (pre_vertex.n_colour_bits <<
                 (BITS_PER_SHORT - N_COLOUR_BITS_BITS)) +
                first_conn_index)
            # ... core mask, mask shift ...
            source_data.append((mask_shift << BITS_PER_SHORT) + core_mask)
            # ... n_dims ...
            source_data.append(n_dims)

            # Add the dimensions; calculations are in reverse order!
            cum_size = 1
            cum_cores_per_dim = 1
            cum_last_size = 1
            all_dim_data = list()
            for i in range(n_dims):
                dim_data = list()
                # Size per core
                dim_data.append(first_slice.shape[i])
                dim_data.append(cum_size)
                dim_data.append(get_div_const(cum_size))
                cum_size *= first_slice.shape[i]

                # Cores
                cores_per_dim = int(ceil(pre_shape[i] / first_slice.shape[i]))
                dim_data.append(cores_per_dim)
                dim_data.append(cum_cores_per_dim)
                dim_data.append(get_div_const(cum_cores_per_dim))
                cum_cores_per_dim *= cores_per_dim

                # Last core
                dim_data.append(last_slice.shape[i])
                dim_data.append(cum_last_size)
                dim_data.append(get_div_const(cum_last_size))
                cum_last_size *= last_slice.shape[i]
                all_dim_data.append(dim_data)
            for dim_data in reversed(all_dim_data):
                source_data.extend(dim_data)

        # Write the spec
        n_post = int(numpy.prod(machine_vertex.vertex_slice.shape))
        spec.write_value(n_post, data_type=DataType.UINT32)
        spec.write_value(len(sources), data_type=DataType.UINT32)
        spec.write_value(n_connectors, data_type=DataType.UINT32)
        spec.write_array(numpy.array(source_data, dtype=numpy.uint32))
        spec.write_array(numpy.concatenate(connector_data))

    def __get_sources_for_target(
            self, app_vertex: PopulationVertex) -> Dict[
            Tuple[ColouredApplicationVertex, str], List[Source]]:
        """
        Get all the application vertex sources that will hit the given
        application vertex.

        :param app_vertex: The vertex being targeted
        :return:
            A dict of source ApplicationVertex and partition id to list of
            source information
        """
        sources = self.__cached_sources.get(app_vertex)
        if sources is None:
            sources = get_sources_for_target(app_vertex)
            self.__cached_sources[app_vertex] = sources
        return sources

    @staticmethod
    def __get_synapse_type(proj: Projection, target: str) -> int:
        edge = proj._projection_edge  # pylint: disable=protected-access
        synapse_type = edge.post_vertex.get_synapse_id_by_target(target)
        # Checked during connection validation, assumed constant
        assert synapse_type is not None
        return synapse_type

    @overrides(AbstractSupportsSignedWeights.get_positive_synapse_index)
    def get_positive_synapse_index(
            self, incoming_projection: Projection) -> int:
        return self.__get_synapse_type(
            incoming_projection,
            self.__connector(incoming_projection).positive_receptor_type)

    @overrides(AbstractSupportsSignedWeights.get_negative_synapse_index)
    def get_negative_synapse_index(
            self, incoming_projection: Projection) -> int:
        return self.__get_synapse_type(
            incoming_projection,
            self.__connector(incoming_projection).negative_receptor_type)

    @overrides(AbstractSupportsSignedWeights.get_maximum_positive_weight)
    def get_maximum_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        # We know the connector doesn't care about the argument
        max_weight = numpy.amax(conn.weights)
        return max_weight if max_weight > 0 else 0

    @overrides(AbstractSupportsSignedWeights.get_minimum_negative_weight)
    def get_minimum_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        # This is different because the connector happens to support this
        min_weight = numpy.amin(conn.weights)
        return min_weight if min_weight < 0 else 0

    @overrides(AbstractSupportsSignedWeights.get_mean_positive_weight)
    def get_mean_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        weights = conn.weights
        if isinstance(weights, (int, float)):
            return weights
        pos_weights = weights[weights > 0]
        if len(pos_weights) == 0:
            return 0
        return numpy.mean(pos_weights)

    @overrides(AbstractSupportsSignedWeights.get_mean_negative_weight)
    def get_mean_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        weights = conn.weights
        if isinstance(weights, (int, float)):
            return weights
        neg_weights = weights[weights < 0]
        if len(neg_weights) == 0:
            return 0
        return numpy.mean(neg_weights)

    @overrides(AbstractSupportsSignedWeights.get_variance_positive_weight)
    def get_variance_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        weights = conn.weights
        if isinstance(weights, (int, float)):
            return 0
        pos_weights = weights[weights > 0]
        if len(pos_weights) == 0:
            return 0
        return numpy.var(pos_weights)

    @overrides(AbstractSupportsSignedWeights.get_variance_negative_weight)
    def get_variance_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        weights = conn.weights
        if isinstance(weights, (int, float)):
            return 0
        neg_weights = weights[weights < 0]
        if len(neg_weights) == 0:
            return 0
        return numpy.var(neg_weights)

    @property
    @overrides(AbstractLocalOnly.synapses_per_second)
    def synapses_per_second(self) -> int:
        # This is a guess!
        return 15000000



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.models.neuron import AbstractPyNNNeuronModelStandard
from spynnaker.pyNN.models.defaults import default_initial_values
from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.models.neuron.neuron_models import (
    NeuronModelLeakyIntegrateAndFire)
from spynnaker.pyNN.models.neuron.synapse_types import (
    SynapseTypeDualExponential)
from spynnaker.pyNN.models.neuron.input_types import InputTypeCurrent
from spynnaker.pyNN.models.neuron.threshold_types import ThresholdTypeStatic


class IFCurrDualExpBase(AbstractPyNNNeuronModelStandard):
    """
    Leaky integrate and fire neuron with two exponentially decaying
    excitatory current inputs, and one exponentially decaying inhibitory
    current input.

    """

    @default_initial_values({"v", "isyn_exc", "isyn_exc2", "isyn_inh"})
    def __init__(
            self, tau_m: ModelParameter = 20.0, cm: ModelParameter = 1.0,
            v_rest: ModelParameter = -65.0, v_reset: ModelParameter = -65.0,
            v_thresh: ModelParameter = -50.0, tau_syn_E: ModelParameter = 5.0,
            tau_syn_E2: ModelParameter = 5.0, tau_syn_I: ModelParameter = 5.0,
            tau_refrac: ModelParameter = 0.1, i_offset: ModelParameter = 0.0,
            v: ModelParameter = -65.0, isyn_exc: ModelParameter = 0.0,
            isyn_inh: ModelParameter = 0.0, isyn_exc2: ModelParameter = 0.0):
        """
        :param tau_m: :math:`\\tau_m`
        :param cm: :math:`C_m`
        :param v_rest: :math:`V_{rest}`
        :param v_reset: :math:`V_{reset}`
        :param v_thresh: :math:`V_{thresh}`
        :param tau_syn_E: :math:`\\tau^{syn}_{e_1}`
        :param tau_syn_E2: :math:`\\tau^{syn}_{e_2}`
        :param tau_syn_I: :math:`\\tau^{syn}_i`
        :param tau_refrac: :math:`\\tau_{refrac}`
        :param i_offset: :math:`I_{offset}`
        :param v: :math:`V_{init}`
        :param isyn_exc: :math:`I^{syn}_{e_1}`
        :param isyn_inh: :math:`I^{syn}_i`
        :param isyn_exc2: :math:`I^{syn}_{e_2}`
        """
        neuron_model = NeuronModelLeakyIntegrateAndFire(
            v, v_rest, tau_m, cm, i_offset, v_reset, tau_refrac)
        synapse_type = SynapseTypeDualExponential(
            tau_syn_E, tau_syn_E2, tau_syn_I, isyn_exc, isyn_exc2, isyn_inh)
        input_type = InputTypeCurrent()
        threshold_type = ThresholdTypeStatic(v_thresh)

        super().__init__(
            model_name="IF_curr_dual_exp", binary="IF_curr_exp_dual.aplx",
            neuron_model=neuron_model, input_type=input_type,
            synapse_type=synapse_type, threshold_type=threshold_type)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, List, Union

import numpy
from numpy import floating
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from pyNN.random import RandomDistribution

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.ranged import RangeDictionary, RangedList

from spynnaker.pyNN.utilities.ranged import SpynnakerRangedList
from spynnaker.pyNN.utilities.struct import Struct

#: The type of parameters to a neuron model.
ModelParameter: TypeAlias = Union[
    float, Iterable[float], RandomDistribution, NDArray[floating]]


class AbstractStandardNeuronComponent(object, metaclass=AbstractBase):
    """
    Represents a component of a standard neural model.
    """
    __slots__ = (
        "__structs",
        "__units")

    def __init__(self, structs: List[Struct], units: Dict[str, str]):
        """
        :param structs: The structures of the component
        :param units: The units to use for each parameter
        """
        self.__structs = structs
        self.__units = units

    @property
    def structs(self) -> List[Struct]:
        """
        The structures of the component.  If there are multiple structures,
        the order is how they will appear in memory; where there are
        structures that repeat per neuron the repeats will appear adjacent
        e.g. for non-repeating structure `g`, followed by repeating structures
        `s1` and `s2` with 3 neurons the layout will be:
        ``[g, s1, s1, s1, s2, s2, s2]``.
        """
        return self.__structs

    @abstractmethod
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        """
        Add the initial values of the parameters to the parameter holder.

        :param parameters: A holder of the parameters
        """
        raise NotImplementedError

    @abstractmethod
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        """
        Add the initial values of the state variables to the state
        variables holder.

        :param state_variables: A holder of the state variables
        """
        raise NotImplementedError

    def has_variable(self, variable: str) -> bool:
        """
        Determine if this component has a variable by the given name.

        :param variable: The name of the variable
        :returns: True if the variable exists, False otherwise
        """
        return variable in self.__units

    def get_units(self, variable: str) -> str:
        """
        :param variable: The name of the variable
        :returns: The units of the given variable.
        """
        return self.__units[variable]

    @staticmethod
    def _convert(value: ModelParameter) -> \
            Union[float, RangedList[float], RandomDistribution]:
        """
        Converts a model parameter into a form that can be ingested by a
        RangeDictionary.
        """
        if isinstance(value, (float, int, numpy.integer, numpy.floating)):
            return float(value)
        if isinstance(value, RandomDistribution):
            return value
        # TODO: Is this correct? Without this, things will only handle floats
        return SpynnakerRangedList(None, value)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Plotting tools to be used together with
https://github.com/NeuralEnsemble/PyNN/blob/master/pyNN/utility/plotting.py
"""

from typing import Any, Dict, Final, List, Union
from types import ModuleType

from neo import SpikeTrain, Block, Segment, AnalogSignal
from neo.core.spiketrainlist import SpikeTrainList  # type: ignore[import]
import numpy as np
from numpy.typing import NDArray
import quantities
from typing_extensions import TypeAlias

plt: ModuleType
try:
    from pyNN.utility.plotting import repeat
    import matplotlib.pyplot
    from matplotlib.axes import Axes
    plt = matplotlib.pyplot
    _matplotlib_missing = False
except ImportError:
    _matplotlib_missing = True

TA_DATA: Final['TypeAlias'] = Union[
    List[SpikeTrain], SpikeTrainList, AnalogSignal, NDArray, Block, Segment]


def _handle_options(axes: Axes, options: Dict[str, Any]) -> None:
    """
    Handles options that can not be passed to `axes.plot`.

    Removes the ones it has handled

    axes.plot will throw an exception if it gets unwanted options

    :param axes: An Axes in a matplotlib figure
    :param options: All options the plotter can be configured with
    """
    if "xticks" not in options or options.pop("xticks") is False:
        plt.setp(axes.get_xticklabels(), visible=False)
    if "xlabel" in options:
        axes.set_xlabel(options.pop("xlabel"))
    else:
        axes.set_xlabel("Time (ms)")
    if "yticks" not in options or options.pop("yticks") is False:
        plt.setp(axes.get_yticklabels(), visible=False)
    if "ylabel" in options:
        axes.set_ylabel(options.pop("ylabel"))
    else:
        axes.set_ylabel("Neuron index")
    if "ylim" in options:
        axes.set_ylim(options.pop("ylim"))
    if "xlim" in options:
        axes.set_xlim(options.pop("xlim"))


def _plot_spikes(axes: Axes, spike_times: NDArray, neurons: NDArray,
                 label: str = '', **options: Any) -> None:
    """
    Plots the spikes based on two lists.

    :param axes: An Axes in a matplotlib figure
    :param spike_times: List of spike times
    :param neurons: List of Neuron IDs
    :param label: Label for the graph
    :param options: plotting options
    """
    if len(neurons):
        max_index = max(neurons)
        min_index = min(neurons)
        axes.plot(spike_times, neurons, 'b.', **options)
        axes.set_ylim(-0.5 + min_index, max_index + 0.5)
    if label:
        plt.text(0.95, 0.95, label,
                 transform=axes.transAxes, ha='right', va='top',
                 bbox=dict(facecolor='white', alpha=1.0))


def plot_spiketrains(
        axes: Axes, spiketrains: Union[List[SpikeTrain], SpikeTrainList],
        label: str = '', **options: Any) -> None:
    """
    Plot all spike trains in a Segment in a raster plot.

    :param axes: An Axes in a matplotlib figure
    :param spiketrains: List of spike times
    :param label: Label for the graph
    :param options: plotting options
    """
    axes.set_xlim(0, spiketrains[0].t_stop / quantities.ms)
    _handle_options(axes, options)
    neurons = np.concatenate(
        [np.repeat(x.annotations['source_index'], len(x))
         for x in spiketrains])
    spike_times = np.concatenate(spiketrains, axis=0)
    _plot_spikes(axes, spike_times, neurons, label=label, **options)


def plot_spikes_numpy(axes: Axes, spikes: NDArray, label: str = '',
                      **options: Any) -> None:
    """
    Plot all spikes.

    :param axes: An Axes in a matplotlib figure
    :param spikes: sPyNNaker7 format numpy array of spikes
    :param label: Label for the graph
    :param options: plotting options
    """
    _handle_options(axes, options)
    neurons = spikes[:, 0]
    spike_times = spikes[:, 1]
    _plot_spikes(axes, spike_times, neurons, label=label, **options)


def _heat_plot(axes: Axes, values: NDArray, label: str = '',
               **options: Any) -> None:
    """
    Plots three lists of neurons, times and values into a heat map.

    :param axes: An Axes in a matplotlib figure
    :param values: List of values to plot
    :param label: Label for the graph
    :param options: plotting options
    """
    _handle_options(axes, options)
    heat_map = axes.imshow(values, cmap='hot', interpolation='none',
                           origin='lower', aspect='auto')
    fig = axes.figure
    assert fig is not None
    fig.colorbar(heat_map)
    if label:
        plt.text(0.95, 0.95, label,
                 transform=axes.transAxes, ha='right', va='top',
                 bbox=dict(facecolor='white', alpha=1.0))


def heat_plot_numpy(axes: Axes, data: NDArray, label: str = '',
                    **options: Any) -> None:
    """
    Plots neurons, times and values into a heat map.

    :param axes: An Axes in a matplotlib figure
    :param data: numpy array of values in spynnaker7 format
    :param label: Label for the graph
    :param options: plotting options
    """
    neurons = data[:, 0].astype(np.uint32)
    times = data[:, 1].astype(np.uint32)
    values = data[:, 2]
    info_array = np.empty((neurons.max() + 1, times.max() + 1))
    info_array[:] = np.nan
    info_array[neurons, times] = values
    _heat_plot(axes, info_array, label=label, **options)


def heat_plot_neo(axes: Axes, signal_array: AnalogSignal, label: str = '',
                  **options: Any) -> None:
    """
    Plots neurons, times and values into a heat map.

    :param axes: An Axes in a matplotlib figure
    :param signal_array: Neo Signal array object
    :param label: Label for the graph
    :param options: plotting options
    """
    if label is None:
        label = signal_array.name
    values = np.transpose(signal_array.magnitude)
    _heat_plot(axes, values, label=label, **options)


def plot_segment(axes: Axes, segment: Segment, label: str = '',
                 **options: Any) -> None:
    """
    Plots a segment into a plot of spikes or a heat map.

    If there is more than ode type of Data in the segment options must
    include the name of the data to plot

    .. note::
        Method signature defined by PyNN plotting.
        This allows mixing of this plotting tool and PyNN's

    :param axes: An Axes in a matplotlib figure
    :param segment: Data for one run to plot
    :param label: Label for the graph
    :param options: plotting options
    """
    analogsignals = segment.analogsignals
    if "name" in options:
        name = options.pop("name")
        if name == 'spikes':
            plot_spiketrains(axes, segment.spiketrains, label=label, **options)
        else:
            heat_plot_neo(
                axes, segment.filter(name=name)[0], label=label, **options)
    elif segment.spiketrains:
        if len(analogsignals) > 1:
            raise ValueError("please specify data to plot using name=")
        plot_spiketrains(axes, segment.spiketrains, label=label, **options)
    elif len(analogsignals) == 1:
        heat_plot_neo(axes, analogsignals[0], label=label, **options)
    elif len(analogsignals) > 1:
        raise ValueError("please specify data to plot using name=")
    else:
        raise ValueError("Block does not appear to hold any data")


class SpynnakerPanel(object):
    """
    Represents a single panel in a multi-panel figure.

    Compatible with :py:class:`pyNN.utility.plotting.Frame` and
    can be mixed with :py:class:`pyNN.utility.plotting.Panel`

    Unlike :py:class:`pyNN.utility.plotting.Panel`,
    Spikes are plotted faster,
    other data is plotted as a heat map.

    A panel is a Matplotlib Axes or Subplot instance. A data item may be an
    :py:class:`~neo.core.AnalogSignal`, or a list of
    :py:class:`~neo.core.SpikeTrain`\\ s. The Panel will
    automatically choose an appropriate representation. Multiple data items
    may be plotted in the same panel.

    Valid options are any valid Matplotlib formatting options that should be
    applied to the Axes/Subplot, plus in addition:

        `data_labels`:
            a list of strings of the same length as the number of data items.
        `line_properties`:
            a list of dictionaries containing Matplotlib formatting options,\
            of the same length as the number of data items.

    Whole Neo Objects can be passed in as long as they
    contain a single Segment/run
    and only contain one type of data.
    Whole Segments can be passed in only if they only contain one type of data.
    """

    def __init__(self, *data: TA_DATA, **options: Any):
        """
        :param data: One or more data series to be plotted.
        :param options: Any additional information.
        """
        if _matplotlib_missing:
            raise ImportError("No matplotlib module found")
        self.data = list(data)
        self.options = options
        self.data_labels = options.pop("data_labels", repeat(None))
        self.line_properties = options.pop("line_properties", repeat({}))

    def plot(self, axes: Axes) -> None:
        """
        Plot the Panel's data in the provided Axes/Subplot instance.

        :param axes: An Axes in a matplotlib figure
        """
        for datum, label, properties in zip(self.data, self.data_labels,
                                            self.line_properties):
            properties.update(self.options)

            # Support lists length one
            # for example result of segments[0].filter(name='v')
            if isinstance(datum, list):
                if not datum:
                    raise ValueError("Can't handle empty list")
                if len(datum) == 1 and not isinstance(datum[0], SpikeTrain):
                    datum = datum[0]

            if isinstance(datum, (list, SpikeTrainList)):
                self.__plot_list(axes, datum, label, properties)
            # AnalogSignal is also a ndarray, but data format different!
            # We import them as a single name here
            elif isinstance(datum, AnalogSignal):
                heat_plot_neo(axes, datum, label=label, **properties)
            elif isinstance(datum, np.ndarray):
                self.__plot_array(axes, datum, label, properties)
            elif isinstance(datum, Block):
                self.__plot_block(axes, datum, label, properties)
            elif isinstance(datum, Segment):
                plot_segment(axes, datum, label=label, **properties)
            else:
                raise ValueError(f"Can't handle type {type(datum)}; "
                                 f"consider using pyNN.utility.plotting")

    @staticmethod
    def __plot_list(
            axes: Axes, datum: Union[List[SpikeTrain], SpikeTrainList],
            label: str, properties: Dict[str, Any]) -> None:
        if not isinstance(datum[0], SpikeTrain):
            raise ValueError(f"Can't handle lists of type {type(datum)}")
        plot_spiketrains(axes, datum, label=label, **properties)

    @staticmethod
    def __plot_array(axes: Axes, datum: NDArray, label: str,
                     properties: Dict[str, Any]) -> None:
        if len(datum[0]) == 2:
            plot_spikes_numpy(axes, datum, label=label, **properties)
        elif len(datum[0]) == 3:
            heat_plot_numpy(axes, datum, label=label, **properties)
        else:
            raise ValueError(
                f"Can't handle ndarray with {len(datum[0])} columns")

    @staticmethod
    def __plot_block(axes: Axes, datum: Block, label: str,
                     properties: Dict[str, Any]) -> None:
        if "run" in properties:
            run = int(properties.pop("run"))
            if len(datum.segments) <= run:
                raise ValueError(
                    f"Block only has {len(datum.segments)} segments")
            segment = datum.segments[run]
        elif len(datum.segments) != 1:
            raise ValueError(f"Block has {len(datum.segments)} segments "
                             "please specify one to plot using run=")
        else:
            segment = datum.segments[0]
        plot_segment(axes, segment, label=label, **properties)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
These are particular configurations of neuron model components. Each of them
is either supported or explicitly not supported.

*Others are possible* but might require an unknown amount of work to get
running on SpiNNaker.
"""

from .eif_cond_alpha_isfa_ista import EIFConductanceAlphaPopulation
from .hh_cond_exp import HHCondExp
from .if_cond_alpha import IFCondAlpha
from .if_cond_exp_base import IFCondExpBase
from .if_curr_alpha import IFCurrAlpha
from .if_curr_dual_exp_base import IFCurrDualExpBase
from .if_curr_exp_base import IFCurrExpBase
from .if_facets_hardware1 import IFFacetsConductancePopulation
from .izk_cond_exp_base import IzkCondExpBase
from .izk_cond_dual_exp_base import IzkCondDualExpBase
from .izk_curr_exp_base import IzkCurrExpBase
from .if_cond_exp_stoc import IFCondExpStoc
from .if_curr_delta import IFCurrDelta
from .if_curr_exp_ca2_adaptive import IFCurrExpCa2Adaptive
from .if_curr_exp_semd_base import IFCurrExpSEMDBase
from .if_curr_delta_ca2_adaptive import IFCurrDeltaCa2Adaptive
from .stoc_exp import StocExp
from .stoc_exp_stable import StocExpStable
from .stoc_sigma import StocSigma
from .if_trunc_delta import IFTruncDelta
from .if_curr_delta_fixed_prob import IFCurrDeltaFixedProb

__all__ = ["EIFConductanceAlphaPopulation", "HHCondExp", "IFCondAlpha",
           "IFCondExpBase", "IFCurrAlpha", "IFCurrDualExpBase",
           "IFCurrExpBase", "IFFacetsConductancePopulation", "IzkCondExpBase",
           "IzkCurrExpBase", "IFCondExpStoc", "IzkCondDualExpBase",
           "IFCurrDelta", "IFCurrExpCa2Adaptive", "IFCurrExpSEMDBase",
           "IFCurrDeltaCa2Adaptive", "StocExp", "StocExpStable", "StocSigma",
           "IFTruncDelta", "IFCurrDeltaFixedProb"]



# Copyright (c) 2025 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pacman.model.graphs.application import ApplicationVertex

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.extra_algorithms.splitter_components import (
    SplitterPoissonDelegate)
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector, OneToOneConnector)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamics, SynapseDynamicsStatic)
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVertex
from spynnaker.pyNN.types import DELAYS


def is_direct_poisson_source(
        post_vertex: ApplicationVertex, pre_vertex: ApplicationVertex,
        connector: AbstractConnector, dynamics: AbstractSynapseDynamics,
        delay: DELAYS) -> bool:
    """
    :param post_vertex: The receiving vertex
    :param pre_vertex: The vertex sending into the Projection
    :param connector: The connector in use in the Projection
    :param dynamics:
        The synapse dynamics in use in the Projection
    :param delay:
        The delay in use in the Projection
    :returns: True if a given Poisson source can be created by this splitter.
    """
    return (isinstance(pre_vertex, SpikeSourcePoissonVertex) and
            isinstance(pre_vertex.splitter, SplitterPoissonDelegate) and
            len(pre_vertex.outgoing_projections) == 1 and
            pre_vertex.n_atoms == post_vertex.n_atoms and
            isinstance(connector, OneToOneConnector) and
            isinstance(dynamics, SynapseDynamicsStatic) and
            delay == SpynnakerDataView().get_simulation_time_step_ms())



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict, Iterable, List, Sequence, Tuple, cast

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides
from spinn_utilities.ordered_set import OrderedSet

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM, MultiRegionSDRAM
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.utilities.utility_objs import ChipCounter

from spynnaker.pyNN.models.common.population_application_vertex import (
    PopulationApplicationVertex)
from spynnaker.pyNN.models.neuron import (
    PopulationMachineVertex,
    PopulationMachineLocalOnlyCombinedVertex, LocalOnlyProvenance)
from spynnaker.pyNN.models.neuron.population_machine_vertex import (
    NeuronProvenance, SynapseProvenance, MainProvenance,
    SpikeProcessingProvenance)
from spynnaker.pyNN.models.neuron.master_pop_table import (
    MasterPopTableAsBinarySearch)
from spynnaker.pyNN.utilities.bit_field_utilities import (
    get_sdram_for_bit_field_region)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamicsStructural)
from spynnaker.pyNN.models.neuron.local_only import AbstractLocalOnly
from spynnaker.pyNN.models.utility_models.delays import DelayExtensionVertex
from spynnaker.pyNN.models.neuron.synaptic_matrices import SynapticMatrices
from spynnaker.pyNN.models.neuron.neuron_data import NeuronData
from spynnaker.pyNN.models.neuron.population_machine_common import (
    PopulationMachineCommon)

from .splitter_population_vertex import SplitterPopulationVertex


class SplitterPopulationVertexFixed(SplitterPopulationVertex):
    """
    Handles the splitting of the :py:class:`PopulationVertex`
    using fixed slices.
    """

    @overrides(AbstractSplitterCommon.create_machine_vertices)
    def create_machine_vertices(self, chip_counter: ChipCounter) -> None:
        app_vertex = self.governed_app_vertex
        app_vertex.synapse_recorder.add_region_offset(
            len(app_vertex.neuron_recorder.get_recordable_variables()))

        max_atoms_per_core = min(
            app_vertex.get_max_atoms_per_core(), app_vertex.n_atoms)

        ring_buffer_shifts = app_vertex.get_ring_buffer_shifts()
        weight_scales = app_vertex.get_weight_scales(ring_buffer_shifts)
        all_syn_block_sz = app_vertex.get_synapses_size(
            max_atoms_per_core)
        structural_sz = app_vertex.get_structural_dynamics_size(
            max_atoms_per_core)
        sdram = self.get_sdram_used_by_atoms(
            max_atoms_per_core, all_syn_block_sz, structural_sz)
        synapse_regions = PopulationMachineVertex.SYNAPSE_REGIONS
        synaptic_matrices = SynapticMatrices(
            app_vertex, synapse_regions, max_atoms_per_core, weight_scales,
            all_syn_block_sz)
        neuron_data = NeuronData(app_vertex)

        for index, vertex_slice in enumerate(self._get_fixed_slices()):
            chip_counter.add_core(sdram)
            label = f"{app_vertex.label}{vertex_slice}"
            machine_vertex = self.create_machine_vertex(
                vertex_slice, sdram, label,
                structural_sz, ring_buffer_shifts, weight_scales,
                index, max_atoms_per_core, synaptic_matrices, neuron_data)
            app_vertex.remember_machine_vertex(machine_vertex)

    @overrides(AbstractSplitterCommon.get_in_coming_slices)
    def get_in_coming_slices(self) -> List[Slice]:
        return self._get_fixed_slices()

    @overrides(AbstractSplitterCommon.get_out_going_slices)
    def get_out_going_slices(self) -> List[Slice]:
        return self._get_fixed_slices()

    @overrides(AbstractSplitterCommon.get_out_going_vertices)
    def get_out_going_vertices(
            self, partition_id: str) -> Sequence[MachineVertex]:
        return list(self.governed_app_vertex.machine_vertices)

    @overrides(AbstractSplitterCommon.get_in_coming_vertices)
    def get_in_coming_vertices(
            self, partition_id: str) -> Sequence[MachineVertex]:
        return list(self.governed_app_vertex.machine_vertices)

    @overrides(AbstractSplitterCommon.get_source_specific_in_coming_vertices)
    def get_source_specific_in_coming_vertices(
            self, source_vertex: ApplicationVertex,
            partition_id: str) -> List[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        # Determine the real pre-vertex
        pre_vertex = source_vertex
        if isinstance(source_vertex, DelayExtensionVertex):
            pre_vertex = source_vertex.source_vertex
        if not isinstance(pre_vertex, PopulationApplicationVertex):
            return []

        # Use the real pre-vertex to get the projections
        targets: Dict[MachineVertex, OrderedSet[
            AbstractVertex]] = defaultdict(OrderedSet)
        for proj in self.governed_app_vertex.get_incoming_projections_from(
                pre_vertex):
            # pylint: disable=protected-access
            s_info = proj._synapse_information
            # Use the original source vertex to get the connected vertices,
            # as the real source machine vertices must make it in to this array
            for (tgt, srcs) in s_info.synapse_dynamics.get_connected_vertices(
                    s_info, source_vertex, self.governed_app_vertex):
                targets[tgt].update(srcs)
        return [(tgt, tuple(srcs)) for tgt, srcs in targets.items()]

    @overrides(AbstractSplitterCommon.machine_vertices_for_recording)
    def machine_vertices_for_recording(
            self, variable_to_record: str) -> Iterable[MachineVertex]:
        return self.governed_app_vertex.machine_vertices

    def create_machine_vertex(
            self, vertex_slice: Slice, sdram: AbstractSDRAM, label: str,
            structural_sz: int, ring_buffer_shifts: Sequence[int],
            weight_scales: NDArray[floating], index: int,
            max_atoms_per_core: int, synaptic_matrices: SynapticMatrices,
            neuron_data: NeuronData) -> PopulationMachineCommon:
        """
        :returns: A new Machine vertex for this slice.
        """
        # If using local-only create a local-only vertex
        s_dynamics = self.governed_app_vertex.synapse_dynamics
        if isinstance(s_dynamics, AbstractLocalOnly):
            return PopulationMachineLocalOnlyCombinedVertex(
                sdram, label, self.governed_app_vertex, vertex_slice, index,
                ring_buffer_shifts, weight_scales, neuron_data,
                max_atoms_per_core)

        # Otherwise create a normal vertex
        return PopulationMachineVertex(
            sdram, label, self.governed_app_vertex, vertex_slice, index,
            ring_buffer_shifts,
            weight_scales, structural_sz, max_atoms_per_core,
            synaptic_matrices, neuron_data)

    def get_sdram_used_by_atoms(
            self, n_atoms: int, all_syn_block_sz: int,
            structural_sz: int) -> AbstractSDRAM:
        """
        :returns: The SDRAM resources of a slice of atoms.
        """
        variable_sdram = self.__get_variable_sdram(n_atoms)
        constant_sdram = self.__get_constant_sdram(
            n_atoms, all_syn_block_sz, structural_sz)
        sdram = MultiRegionSDRAM()
        sdram.nest(len(PopulationMachineVertex.REGIONS) + 1, variable_sdram)
        sdram.merge(constant_sdram)

        # return the total resources.
        return sdram

    def __get_variable_sdram(self, n_atoms: int) -> AbstractSDRAM:
        """
        Returns the variable SDRAM from the recorders.

        :param n_atoms: The number of atoms to account for
        :return: the variable SDRAM used by the neuron recorder
        """
        s_dynamics = self.governed_app_vertex.synapse_dynamics
        if isinstance(s_dynamics, AbstractSynapseDynamicsStructural):
            max_rewires_per_ts = s_dynamics.get_max_rewires_per_ts()
            self.governed_app_vertex.synapse_recorder.set_max_rewires_per_ts(
                max_rewires_per_ts)

        return (
            self.governed_app_vertex.get_max_neuron_variable_sdram(n_atoms) +
            self.governed_app_vertex.get_max_synapse_variable_sdram(n_atoms))

    def __get_constant_sdram(
            self, n_atoms: int, all_syn_block_sz: int,
            structural_sz: int) -> MultiRegionSDRAM:
        """
        Returns the constant SDRAM used by the atoms.

        :param n_atoms: The number of atoms to account for
        """
        s_dynamics = self.governed_app_vertex.synapse_dynamics
        n_record = (
            len(self.governed_app_vertex.neuron_recordables) +
            len(self.governed_app_vertex.synapse_recordables))

        n_provenance = NeuronProvenance.N_ITEMS + MainProvenance.N_ITEMS
        if isinstance(s_dynamics, AbstractLocalOnly):
            n_provenance += LocalOnlyProvenance.N_ITEMS
        else:
            n_provenance += (
                SynapseProvenance.N_ITEMS + SpikeProcessingProvenance.N_ITEMS)

        sdram = MultiRegionSDRAM()
        if isinstance(s_dynamics, AbstractLocalOnly):
            sdram.merge(self.governed_app_vertex.get_common_constant_sdram(
                n_record, n_provenance,
                PopulationMachineLocalOnlyCombinedVertex.COMMON_REGIONS))
            sdram.merge(self.governed_app_vertex.get_neuron_constant_sdram(
                n_atoms,
                PopulationMachineLocalOnlyCombinedVertex.NEURON_REGIONS))
            sdram.merge(self.__get_local_only_constant_sdram(n_atoms))
        else:
            sdram.merge(self.governed_app_vertex.get_common_constant_sdram(
                n_record, n_provenance,
                PopulationMachineVertex.COMMON_REGIONS))
            sdram.merge(self.governed_app_vertex.get_neuron_constant_sdram(
                n_atoms, PopulationMachineVertex.NEURON_REGIONS))
            sdram.merge(self.__get_synapse_constant_sdram(
                n_atoms, all_syn_block_sz, structural_sz))
        return sdram

    def __get_local_only_constant_sdram(
            self, n_atoms: int) -> MultiRegionSDRAM:
        s_dynamics = cast(AbstractLocalOnly,
                          self.governed_app_vertex.synapse_dynamics)
        sdram = MultiRegionSDRAM()
        sdram.add_cost(
            PopulationMachineLocalOnlyCombinedVertex.REGIONS.LOCAL_ONLY,
            PopulationMachineLocalOnlyCombinedVertex.LOCAL_ONLY_SIZE)
        sdram.add_cost(
            PopulationMachineLocalOnlyCombinedVertex.REGIONS.LOCAL_ONLY_PARAMS,
            s_dynamics.get_parameters_usage_in_bytes(
                n_atoms, self.governed_app_vertex.incoming_projections))
        return sdram

    def __get_synapse_constant_sdram(
            self, n_atoms: int, all_syn_block_sz: int,
            structural_sz: int) -> MultiRegionSDRAM:
        """
        Get the amount of fixed SDRAM used by synapse parts.

        :param n_atoms: The number of atoms to account for
        """
        regions = PopulationMachineVertex.SYNAPSE_REGIONS
        sdram = MultiRegionSDRAM()
        sdram.add_cost(regions.synapse_params,
                       self.governed_app_vertex.get_synapse_params_size())
        sdram.add_cost(regions.synapse_dynamics,
                       self.governed_app_vertex.get_synapse_dynamics_size(
                           n_atoms))
        sdram.add_cost(regions.structural_dynamics, structural_sz)
        sdram.add_cost(regions.synaptic_matrix, all_syn_block_sz)
        sdram.add_cost(
            regions.pop_table,
            MasterPopTableAsBinarySearch.get_master_population_table_size(
                self.governed_app_vertex.incoming_projections))
        sdram.add_cost(regions.connection_builder,
                       self.governed_app_vertex.get_synapse_expander_size())
        sdram.add_cost(regions.bitfield_filter,
                       get_sdram_for_bit_field_region(
                           self.governed_app_vertex.incoming_projections))
        return sdram



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from threading import Thread, RLock
from time import sleep
from typing import Any, List

from matplotlib import pyplot
import numpy

from spinn_utilities.log import FormatAdapter

import spynnaker.pyNN.external_devices as external_devices
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotRetinaResolution)
from spynnaker.pyNN.connections import SpynnakerLiveSpikesConnection

_logger = FormatAdapter(logging.getLogger(__name__))
MAX_VALUE = 33.0
ADD_VALUE = 1.0
DECAY_FACTOR = 0.5
SLEEP_TIME = 0.1


class PushBotRetinaViewer():
    """
    Viewer of retina from the PushBot.
    """
    __slots__ = (
        "__image_data", "__image_lock",
        "__without_polarity_mask", "__height",
        "__fig", "__plot",
        "__running", "__conn")

    def __init__(self, retina_resolution: PushBotRetinaResolution,
                 label: str, sim: None = None):
        # pylint: disable=wrong-spelling-in-docstring
        """
        :param retina_resolution: Size of the retina to use
        :param label:
            Label for connection over which live spikes will be received.
        :param sim: Deprecated! Do not use.
        """
        if sim is not None:
            _logger.warning("PushBotRetinaViewer: sim=None is deprecated")
        pyplot.ion()
        self.__image_data = numpy.zeros(
            (retina_resolution.value.pixels, retina_resolution.value.pixels),
            dtype=numpy.float32)
        self.__fig, axes = pyplot.subplots(figsize=(8, 8))
        self.__plot = axes.imshow(
            self.__image_data, interpolation="nearest", cmap="Greens",
            vmin=0, vmax=MAX_VALUE)
        self.__fig.canvas.draw()
        self.__fig.canvas.flush_events()

        self.__without_polarity_mask = (
            2 ** (retina_resolution.value.bits_per_coordinate * 2)) - 1
        self.__height = retina_resolution.value.pixels

        self.__running = True
        self.__image_lock = RLock()

        self.__conn = SpynnakerLiveSpikesConnection(
            receive_labels=[label], local_port=None)
        self.__conn.add_receive_callback(label, self.__recv)

    @property
    def port(self) -> int:
        """
        The port the connection is listening on.
        """
        return self.__conn.local_port

    def __recv(self, label: str, time: int, spikes: List[int]) -> None:
        _ = (label, time)
        np_spikes = numpy.array(spikes) & self.__without_polarity_mask
        x_vals, y_vals = numpy.divmod(np_spikes, self.__height)
        self.__image_lock.acquire()
        self.__image_data[x_vals, y_vals] += 1.0
        self.__image_lock.release()

    def __run_sim_forever(self) -> None:
        # UGLY but needed to avoid circular import
        # pylint: disable=import-outside-toplevel
        from pyNN.spiNNaker import end
        try:
            external_devices.run_forever()
            self.__running = False
            end()
        except KeyboardInterrupt:
            pass
        except Exception:  # pylint: disable=broad-except
            _logger.exception("unexpected exception in simulation thread")

    def __run_sim(self, run_time: float) -> None:
        # UGLY but needed to avoid circular import
        # pylint: disable=import-outside-toplevel
        from pyNN.spiNNaker import end, run
        try:
            run(run_time)
            self.__running = False
            end()
        except KeyboardInterrupt:
            pass
        except Exception:  # pylint: disable=broad-except
            _logger.exception("unexpected exception in simulation thread")

    def __run(self) -> None:
        try:
            while self.__running and self.__fig.get_visible():
                self.__image_lock.acquire()
                self.__plot.set_array(self.__image_data)
                self.__fig.canvas.draw()
                self.__fig.canvas.flush_events()
                self.__image_data *= DECAY_FACTOR
                self.__image_lock.release()
                sleep(0.1)
        except KeyboardInterrupt:
            pass
        except Exception:  # pylint: disable=broad-except
            _logger.exception("unexpected exception in drawing thread")

    def __on_close(self, event: Any) -> None:
        _ = event
        self.__running = False

    def run_until_closed(self) -> None:
        """
        Run the viewer and simulation until the viewer is closed.
        """
        run_thread = Thread(target=self.__run_sim_forever)
        run_thread.start()
        try:
            self.__fig.canvas.mpl_connect('close_event', self.__on_close)
            self.__run()
        finally:
            external_devices.request_stop()
            run_thread.join()

    def run(self, run_time: float) -> None:
        """
        Run the viewer and simulation for a fixed time.
        """
        run_thread = Thread(target=self.__run_sim, args=[run_time])
        run_thread.start()
        try:
            self.__run()
        finally:
            run_thread.join()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinn_utilities.overrides import overrides
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from .spike_injector_vertex import SpikeInjectorVertex

_population_parameters = {
    "port": None,
    "virtual_key": None,
    "reserve_reverse_ip_tag": False,
    "splitter": None
}


class SpikeInjector(AbstractPyNNModel):
    """
    Model that creates a Spike Injector Vertex
    """
    __slots__ = ()

    default_population_parameters = _population_parameters

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *,
            port: Optional[int] = None, virtual_key: Optional[int] = None,
            reserve_reverse_ip_tag: bool = False,
            splitter: Optional[AbstractSplitterCommon] = None) \
            -> SpikeInjectorVertex:
        """
        :param port:
        :param virtual_key:
        :param reserve_reverse_ip_tag:
        :param splitter:
        """
        max_atoms_per_core = self.get_model_max_atoms_per_dimension_per_core()
        return SpikeInjectorVertex(
            n_neurons, label, port, virtual_key,
            reserve_reverse_ip_tag, splitter, max_atoms_per_core)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script installs sPyNNaker so that it usable as the ``pyNN.spiNNaker``
module.

.. note::
    This *modifies* your pyNN installation!
"""

import os
import shutil
from types import ModuleType
from packaging.version import Version
import pyNN
import spynnaker.pyNN as sim


# The version of PyNN that we really want
_TARGET_PYNN_VERSION = "0.9"


def version_satisfies(module: ModuleType, requirement: str) -> bool:
    """
    Perform a version check. This code could be smarter...

    :param module:
    :param requirement:
    :return: Whether the module's version satisfies the given requirement
    """
    return Version(module.__version__) >= Version(requirement)


def install_spynnaker_into(module: ModuleType) -> None:
    """
    Do the actual installation by creating a package within the given
    module's implementation. This is very nasty!

    :param module:
    """
    _file = module.__file__
    assert _file is not None
    spinnaker_dir = os.path.join(os.path.dirname(_file), "spiNNaker")
    if not os.path.exists(spinnaker_dir):
        os.mkdir(spinnaker_dir)

    pynn_init = os.path.join(spinnaker_dir, "__init__.py")
    spynnaker_init = os.path.abspath(sim.__file__)
    shutil.copyfile(spynnaker_init, pynn_init)
    print(f"Updated {pynn_init} to be the same as spynnaker.pyNN")


def setup_pynn() -> None:
    """
    Checks pyNN version and creates the spynnaker model in pynn.
    """
    # Check the version and blow up if it isn't there
    if not version_satisfies(pyNN, _TARGET_PYNN_VERSION):
        raise NotImplementedError(
            f"PyNN version {pyNN.__version__} found; "
            f"sPyNNaker requires PyNN version {_TARGET_PYNN_VERSION}")

    # Perform the installation unless we're on READTHEDOCS
    if os.environ.get('READTHEDOCS', 'False') != 'True':
        install_spynnaker_into(pyNN)


if __name__ == "__main__":
    setup_pynn()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_standard_neuron_component import (
    AbstractStandardNeuronComponent, ModelParameter)
from .abstract_neuron_impl import AbstractNeuronImpl
from .neuron_impl_standard import NeuronImplStandard
from .neuron_impl_stoc_exp import NeuronImplStocExp
from .neuron_impl_stoc_exp_stable import NeuronImplStocExpStable
from .neuron_impl_stoc_sigma import NeuronImplStocSigma

__all__ = [
    "AbstractNeuronImpl", "AbstractStandardNeuronComponent",
    "ModelParameter", "NeuronImplStandard", "NeuronImplStocExp",
    "NeuronImplStocExpStable", "NeuronImplStocSigma"]



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from spynnaker.pyNN.models.neuron import AbstractPyNNNeuronModel
from spynnaker.pyNN.models.defaults import default_parameters
from spynnaker.pyNN.models.neuron.implementations import (
    ModelParameter, NeuronImplStocSigma)


class StocSigma(AbstractPyNNNeuronModel):
    """ Stochastic model with sigma threshold and instantaneous synapses.
    """

    @default_parameters({"tau_refrac", "alpha", "bias"})
    def __init__(self, tau_refrac: ModelParameter = 1,
                 alpha: ModelParameter = 1.0, bias: ModelParameter = 0,
                 refract_init: ModelParameter = 0,
                 seed: Optional[int] = None):
        """
        :param tau_refrac:
        :param alpha:
        :param bias:
        :param refract_init:
        :param seed: Seed for the random distribution
        """
        super().__init__(NeuronImplStocSigma(tau_refrac, alpha, bias,
                                             refract_init, seed))



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex
from spinn_front_end_common.utility_models import MultiCastCommand
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.utilities.constants import SPIKE_PARTITION_ID
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotRetinaDevice)
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotRetinaResolution)
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol


class DelayedPayloadMultiCastCommand(MultiCastCommand):
    """
    A Hack to get the key after the zone allocator is run
    """

    def __init__(self, key: int, vertex: 'PushBotSpiNNakerLinkRetinaDevice'):
        """
        :param key:  The key of the command
        :param vertex: Vertex to get the payload from
        """
        super().__init__(key)
        self._vertex = vertex

    @property
    @overrides(MultiCastCommand.payload)
    def payload(self) -> Optional[int]:
        if self._payload is None:
            self._payload = self._vertex.new_key_command_payload()
        return self._payload

    @property
    @overrides(MultiCastCommand.is_payload)
    def is_payload(self) -> bool:
        return self.payload is not None


class PushBotSpiNNakerLinkRetinaDevice(
        AbstractPushBotRetinaDevice, ApplicationSpiNNakerLinkVertex,
        PopulationApplicationVertex):
    """
    Implementation of a PushBot Retina vertex over a Spinnaker link
    """
    __slots__ = ("__new_key_command", )

    def __init__(
            self, spinnaker_link_id: int,
            protocol: MunichIoSpiNNakerLinkProtocol,
            resolution: PushBotRetinaResolution,
            board_address: Optional[str] = None,
            label: Optional[str] = None,
            n_machine_vertices: int = 1):
        """
        :param spinnaker_link_id:
        :param protocol:
        :param resolution:
        :param board_address:
        :param label:
        :param n_machine_vertices:
        """
        super().__init__(protocol, resolution)
        ApplicationSpiNNakerLinkVertex.__init__(
            self, spinnaker_link_id=spinnaker_link_id,
            n_atoms=resolution.value.n_neurons,
            board_address=board_address, label=label,
            n_machine_vertices=n_machine_vertices)

        # stores for the injection aspects
        self.__new_key_command: Optional[DelayedPayloadMultiCastCommand] = \
            None

    def new_key_command_payload(self) -> int:
        """
        Support method to obtain the key after the key allocator has run

        :return: the key
        """
        routing_info = SpynnakerDataView.get_routing_infos()
        key = routing_info.get_key_from(
            self, SPIKE_PARTITION_ID)
        return key

    @property
    @overrides(AbstractPushBotRetinaDevice.start_resume_commands)
    def start_resume_commands(
            self) -> List[MultiCastCommand]:
        # Update the commands with the additional one to set the key
        new_commands: List[MultiCastCommand] = list()
        for command in super().start_resume_commands:
            if command.key == self._protocol.set_retina_transmission_key:
                # This has to be stored so that the payload can be updated
                self.__new_key_command = DelayedPayloadMultiCastCommand(
                    self._protocol.set_retina_key_key, self)
                new_commands.append(self.__new_key_command)
            new_commands.append(command)
        return new_commands



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex

from spynnaker.pyNN.external_devices_models.push_bot.ethernet import (
    PushBotEthernetSpeakerDevice)
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotSpeaker)
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol


class PushBotSpiNNakerLinkSpeakerDevice(
        PushBotEthernetSpeakerDevice, ApplicationSpiNNakerLinkVertex,
        PopulationApplicationVertex):
    """
    The speaker of a PushBot.
    """
    __slots__ = ()

    def __init__(
            self, speaker: PushBotSpeaker,
            protocol: MunichIoSpiNNakerLinkProtocol, spinnaker_link_id: int,
            n_neurons: int = 1,
            label: Optional[str] = None,
            board_address: Optional[str] = None,
            start_active_time: int = 50,
            start_total_period: int = 100,
            start_frequency: Optional[int] = None,
            start_melody: Optional[int] = None):
        """
        :param speaker: Which speaker device to control
        :param protocol: The protocol instance to get commands from
        :param spinnaker_link_id: The SpiNNakerLink connected to
        :param n_neurons: The number of neurons in the device
        :param label: The label of the device
        :param board_address:
            The IP address of the board that the device is connected to
        :param start_active_time: The "active time" to set at the start
        :param start_total_period: The "total period" to set at the start
        :param start_frequency: The "frequency" to set at the start
        :param start_melody: The "melody" to set at the start
        """
        super().__init__(
            speaker, protocol, start_active_time, start_total_period,
            start_frequency, start_melody)
        ApplicationSpiNNakerLinkVertex.__init__(
            self, spinnaker_link_id=spinnaker_link_id, n_atoms=n_neurons,
            board_address=board_address, label=label)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Optional, Sequence, cast, TYPE_CHECKING

import numpy
from numpy import uint32
from numpy.typing import NDArray

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractGenerateConnectorOnMachine)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractGenerateOnMachine)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo

# Address to indicate that the synaptic region is unused
SYN_REGION_UNUSED = 0xFFFFFFFF


class GeneratorData(object):
    """
    Data for each connection of the synapse generator.
    """
    __slots__ = ("__data", )

    BASE_SIZE = 11 * BYTES_PER_WORD

    def __init__(
            self, synaptic_matrix_offset: Optional[int],
            delayed_synaptic_matrix_offset: Optional[int],
            app_edge: ProjectionApplicationEdge,
            synapse_information: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int, max_post_atoms_per_core: int):
        """
        :param synaptic_matrix_offset:
        :param delayed_synaptic_matrix_offset:
        :param app_edge:
        :param synapse_information:
        :param max_row_info:
        :param max_pre_atoms_per_core:
        :param max_post_atoms_per_core:
        """
        # Offsets are used in words in the generator, but only
        # if the values are valid
        if synaptic_matrix_offset is not None:
            synaptic_matrix_offset //= BYTES_PER_WORD
        else:
            synaptic_matrix_offset = SYN_REGION_UNUSED
        if delayed_synaptic_matrix_offset is not None:
            delayed_synaptic_matrix_offset //= BYTES_PER_WORD
        else:
            delayed_synaptic_matrix_offset = SYN_REGION_UNUSED

        # Take care of Population views
        # pylint: disable=protected-access
        pre_lo, pre_hi = synapse_information.pre_population._view_range
        post_lo, post_hi = synapse_information.post_population._view_range

        # Get objects needed for the next bit
        connector = cast(AbstractGenerateConnectorOnMachine,
                         synapse_information.connector)
        synapse_dynamics = cast(AbstractGenerateOnMachine,
                                synapse_information.synapse_dynamics)

        # Create the data needed
        self.__data = list()
        self.__data.append(numpy.array([
            pre_lo, pre_hi, post_lo, post_hi,
            synapse_information.synapse_type,
            synapse_dynamics.gen_matrix_id,
            connector.gen_connector_id,
            connector.gen_weights_id(synapse_information.weights),
            connector.gen_delays_id(synapse_information.delays)
            ], dtype=numpy.uint32))
        self.__data.append(synapse_dynamics.gen_matrix_params(
            synaptic_matrix_offset, delayed_synaptic_matrix_offset, app_edge,
            synapse_information, max_row_info, max_pre_atoms_per_core,
            max_post_atoms_per_core))
        self.__data.append(connector.gen_connector_params(synapse_information))
        self.__data.append(connector.gen_weights_params(
            synapse_information.weights))
        self.__data.append(connector.gen_delay_params(
            synapse_information.delays))

    @property
    def size(self) -> int:
        """
        The size of the generated data, in bytes.
        """
        return sum(len(i) for i in self.__data) * BYTES_PER_WORD

    @property
    def gen_data(self) -> Sequence[NDArray[uint32]]:
        """
        The data to be written for this connection.
        """
        return self.__data



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from typing import (Collection, Dict, List, Optional, Tuple, Union)
import numpy
from numpy.typing import NDArray
from pyNN.space import Grid2D, Grid3D, BaseStructure
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from pacman.model.graphs.common import Slice
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from spinn_front_end_common.utility_models import ReverseIpTagMultiCastSource
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.common import EIEIOSpikeRecorder
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType
from spynnaker.pyNN.models.abstract_models import SupportsStructure
from spynnaker.pyNN.models.common import PopulationApplicationVertex


logger = FormatAdapter(logging.getLogger(__name__))


class SpikeInjectorVertex(
        ReverseIpTagMultiCastSource, PopulationApplicationVertex,
        SupportsStructure):
    """
    An Injector of Spikes for PyNN populations.  This only allows the user
    to specify the virtual_key of the population to identify the population.
    """
    __slots__ = (
        "__spike_recorder",
        "__structure")

    default_parameters = {
        'label': "spikeInjector", 'port': None, 'virtual_key': None}

    SPIKE_RECORDING_REGION_ID = 0

    def __init__(
            self, n_neurons: int, label: str,
            port: Optional[int], virtual_key: Optional[int],
            reserve_reverse_ip_tag: bool,
            splitter: Optional[AbstractSplitterCommon],
            max_atoms_per_core: Optional[
                Union[int, Tuple[int, ...]]] = sys.maxsize):
        """
        :param label: The optional name of the vertex.
        """
        super().__init__(
            n_keys=n_neurons, label=label, receive_port=port,
            virtual_key=virtual_key,
            reserve_reverse_ip_tag=reserve_reverse_ip_tag,
            splitter=splitter, max_atoms_per_core=max_atoms_per_core)

        # Set up for recording
        self.__spike_recorder = EIEIOSpikeRecorder()
        self.__structure: Optional[BaseStructure] = None

    @overrides(SupportsStructure.set_structure)
    def set_structure(self, structure: BaseStructure) -> None:
        self.__structure = structure

    @property
    @overrides(PopulationApplicationVertex.atoms_shape)
    def atoms_shape(self) -> Tuple[int, ...]:
        if isinstance(self.__structure, (Grid2D, Grid3D)):
            return self.__structure.calculate_size(self.n_atoms)
        return super().atoms_shape

    @overrides(PopulationApplicationVertex.get_recordable_variables)
    def get_recordable_variables(self) -> List[str]:
        return ["spikes"]

    @overrides(PopulationApplicationVertex.set_recording)
    def set_recording(
            self, name: str, sampling_interval: Optional[float] = None,
            indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if sampling_interval is not None:
            logger.warning("Sampling interval currently not supported for "
                           "SpikeInjector so being ignored")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeInjector so being ignored")
        self.enable_recording(True)
        self.__spike_recorder.record = True

    @overrides(PopulationApplicationVertex.get_recording_variables)
    def get_recording_variables(self) -> List[str]:
        if self.__spike_recorder.record:
            return ["spikes"]
        return []

    @overrides(PopulationApplicationVertex.set_not_recording)
    def set_not_recording(self, name: str,
                          indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeSourceArray so being ignored")
        self.enable_recording(False)
        self.__spike_recorder.record = False

    @overrides(PopulationApplicationVertex.get_sampling_interval_ms)
    def get_sampling_interval_ms(self, name: str) -> float:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return SpynnakerDataView.get_simulation_time_step_us()

    @overrides(PopulationApplicationVertex.get_data_type)
    def get_data_type(self, name: str) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return None

    @overrides(PopulationApplicationVertex.get_buffer_data_type)
    def get_buffer_data_type(self, name: str) -> BufferDataType:
        if name == "spikes":
            return BufferDataType.EIEIO_SPIKES
        raise KeyError(f"Cannot record {name}")

    @overrides(PopulationApplicationVertex.get_units)
    def get_units(self, name: str) -> str:
        if name == "spikes":
            return ""
        raise KeyError(f"Cannot record {name}")

    @overrides(PopulationApplicationVertex.get_recording_region)
    def get_recording_region(self, name: str) -> int:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return 0

    @overrides(PopulationApplicationVertex.get_neurons_recording)
    def get_neurons_recording(
            self, name: str, vertex_slice: Slice) -> NDArray[numpy.integer]:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return vertex_slice.get_raster_ids()

    def describe(self) -> Dict[str, Union[str, Collection[str]]]:
        """
        :returns: A human-readable description of the cell or synapse type.
        """
        context = {
            "name": "SpikeInjector",
            "default_parameters":
                self.default_parameters,
            "default_initial_values":
                self.default_parameters,
            "parameters": {
                "port": self._eieio_params.receive_port,
                "virtual_key": self._eieio_params.virtual_key},
        }
        return context



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Final, TextIO, Tuple, cast

from spinn_utilities.config_holder import get_report_path
from spinn_utilities.log import FormatAdapter

from spinn_front_end_common.interface.provenance import (
    ProvenanceReader, ProvenanceWriter)

from spynnaker.pyNN.models.neuron import PopulationMachineVertex

logger = FormatAdapter(logging.getLogger(__name__))

#: How to make the redundancy information table in the provenance DB
REDUNDANCY_BY_CORE: Final = f"""
    CREATE VIEW IF NOT EXISTS redundancy_by_core AS
    SELECT pro.x, pro.y, pro.p, core_name,
        received, filtered, invalid, failed,
        filtered + invalid + failed AS redundant,
        filtered + invalid + failed + received AS total,
        (filtered + invalid + failed) * 100.0 /
            (filtered + invalid + failed + received) AS percent
    FROM
        (SELECT x, y, p, total as received, core_name
            FROM core_stats_view
            WHERE description = '{PopulationMachineVertex.SPIKES_PROCESSED}')
        AS pro
    LEFT JOIN
        (SELECT x, y, p, total AS filtered
            FROM core_stats_view
            WHERE description =
                '{PopulationMachineVertex.BIT_FIELD_FILTERED_PACKETS}')
        AS bit
    LEFT JOIN
       (SELECT x, y, p, total AS invalid
           FROM core_stats_view
           WHERE description =
               '{PopulationMachineVertex.INVALID_MASTER_POP_HITS}')
        AS inv
    LEFT JOIN
        (SELECT x, y, p, total AS failed
            FROM core_stats_view
            WHERE description = '{PopulationMachineVertex.GHOST_SEARCHES}')
        AS fai
    WHERE pro.x = bit.x AND pro.y = bit.y AND pro.p = bit.p
        AND pro.x = inv.x AND pro.y = inv.y AND pro.p = inv.p
        AND pro.x = fai.x AND pro.y = fai.y AND pro.p = fai.p
    """


#: How to generate summary information from the redundancy table
REDUNDANCY_SUMMARY: Final = """
    CREATE VIEW IF NOT EXISTS redundancy_summary AS
    SELECT SUM(total), MAX(total), MIN(total), AVG(total),
        SUM(redundant), MAX(redundant), MIN(redundant), AVG(redundant),
        MAX(percent), MIN(percent), AVG(percent),
        SUM(redundant) * 100.0 / SUM(total) as global_percent
    FROM redundancy_by_core
    """


_N_PROV_ITEMS_NEEDED = 4
_MAX = 100


def redundant_packet_count_report() -> None:
    """
    Writes a report detailing the redundant packet counts.
    """
    file_name = get_report_path("path_redundant_packet_count_report")

    try:
        _create_views()
        with open(file_name, "w", encoding="utf-8") as f:
            _write_report(f)
    except Exception as e:  # pylint: disable=broad-except
        logger.exception(
            "Error {} doing redundant_packet_count_report {}:", e, file_name)


def _create_views() -> None:
    with ProvenanceWriter() as db:
        db.cursor().execute(REDUNDANCY_BY_CORE)
        db.cursor().execute(REDUNDANCY_SUMMARY)


def _write_report(output: TextIO) -> None:
    with ProvenanceReader() as db:
        for data in db.run_query("SELECT * FROM redundancy_by_core"):
            (_, _, _, source, _, filtered, invalid, _, redundant, total,
             percent) = cast(Tuple[int, ...], data)
            output.write(f"\ncore {source} \n")
            output.write(f"    {total} packets received.\n")
            output.write(
                f"    {redundant} were detected as redundant packets by the "
                "bitfield filter.\n")
            output.write(
                f"    {filtered} were detected as having no targets "
                "after the DMA stage.\n")
            output.write(
                f"    {invalid} were detected as packets which "
                "we should not have received in the first place.\n")
            output.write(
                "    Overall this makes a redundant percentage of "
                f"{percent}\n")
        for data in db.run_query("SELECT * FROM redundancy_summary LIMIT 1"):
            (sum_total, max_total, min_total, avg_total,
                sum_redundant, max_redundant, min_redundant, avg_redundant,
                max_percent, min_percent, avg_percent, global_percent) = cast(
                    Tuple[float, ...], data)
            output.write(
                f"\nThe total packets flown in system was {sum_total}.\n")
            output.write(
                f"    The max, min and average per core was {max_total}, "
                f"{min_total}, {avg_total} packets.\n")
            output.write(
                f"The total redundant packets was {sum_redundant}.\n")
            output.write(
                f"    The max, min and average per core was {max_redundant}, "
                f"{min_redundant}, {avg_redundant} packets.\n")
            output.write(
                "The percentages of redundant packets was "
                f"{global_percent}.\n")
            output.write(
                "    The max and min percentages of redundant packets are"
                f"{max_percent} and {min_percent}.\n")
            output.write(
                "    The average redundant percentages from each core were "
                f"{avg_percent} accordingly.\n")
            break


if __name__ == '__main__':
    print(REDUNDANCY_BY_CORE)
    print(REDUNDANCY_SUMMARY)



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from numpy import floating
from numpy.typing import NDArray
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.model.graphs import AbstractSupportsSDRAMEdges
from spinn_front_end_common.utilities.constants import BYTES_PER_SHORT


class ReceivesSynapticInputsOverSDRAM(
        AbstractSupportsSDRAMEdges, metaclass=AbstractBase):
    """
    An object that receives synaptic inputs over SDRAM.

    The number of neurons to be sent per synapse type is rounded up to be
    a power of 2.  A sender must send N_BYTES_PER_INPUT of data for each
    synapse type for each neuron, formatted as all the data for each neuron
    for the first synapse type, followed by all the data for each neuron
    for the second, and so on for each synapse type.  Each input is an
    accumulated weight value for the timestep, scaled with the given weight
    scales.
    """

    # The size of each input in bytes
    N_BYTES_PER_INPUT = BYTES_PER_SHORT

    @property
    @abstractmethod
    def weight_scales(self) -> NDArray[floating]:
        """
        A list of scale factors to be applied to weights that get passed
        over SDRAM, one for each synapse type.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def n_bytes_for_transfer(self) -> int:
        """
        The number of bytes to be sent over the channel.  This will be
        calculated using the above numbers, but also rounded up to a number
        of words, and with the number of neurons rounded up to a power of 2.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Mapping
import numpy

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.utilities import utility_calls

from .abstract_current_source import (
    AbstractCurrentSource, CurrentSourceIDs, CurrentParameter)


class NoisyCurrentSource(AbstractCurrentSource):
    """
    A noisy current source beginning at "start" and ending at "stop", with
    noise simulated based on the given mean and standard deviation, and
    updating every `dt` (`dt` should default to the machine time step).
    """
    __slots__ = (
        "__mean",
        "__stdev",
        "__start",
        "__stop",
        "__dt",
        "__rng",
        "__parameters",
        "__parameter_types")

    def __init__(self, mean: float = 0.0, stdev: float = 0.0,
                 start: float = 0.0, stop: float = 0.0, dt: float = 1.0,
                 rng: None = None) -> None:
        """
        :param mean:
        :param stdev:
        :param start:
        :param stop:
        :param dt:
        :param rng:
        """
        # There's probably no need to actually store these as you can't
        # access them directly in pynn anyway
        time_convert_ms = SpynnakerDataView.get_simulation_time_step_per_ms()
        self.__mean = mean
        self.__stdev = stdev
        self.__start = int(start * time_convert_ms)
        self.__stop = int(stop * time_convert_ms)
        self.__dt = dt * time_convert_ms
        if rng is None:
            seed = None
            self.__rng = numpy.random.RandomState(seed)
        # TODO: What happens if we pass a non-None rng?

        # Error if dt is not the same as machine time step
        if dt != (1 / time_convert_ms):
            raise SpynnakerException(
                "Only currently supported for dt = machine_time_step, here "
                f"dt = {dt} and machine_time_step = {1 / time_convert_ms}")

        self.__parameter_types = {
            'mean': DataType.S1615,
            'stdev': DataType.S1615,
            'start': DataType.UINT32,
            'stop': DataType.UINT32,
            'dt': DataType.S1615,
            'seed': DataType.UINT32}

        self.__parameters: Dict[str, CurrentParameter] = {
            'mean': self.__mean,
            'stdev': self.__stdev,
            'start': self.__start,
            'stop': self.__stop,
            'dt': self.__dt,
            'seed': utility_calls.create_mars_kiss_seeds(self.__rng)}

        super().__init__()

    @overrides(AbstractCurrentSource.set_parameters)
    def set_parameters(self, **parameters: CurrentParameter) -> None:
        for key, value in parameters.items():
            if key not in self.__parameters:
                # throw an exception
                raise SpynnakerException(f"{key} is not a parameter of {self}")
            self.__parameters[key] = value

        # Parameters have been set, so if multi-run then it will have been
        # injected already; if not then it can just be ignored
        if self.app_vertex is not None:
            for m_vertex in self.app_vertex.machine_vertices:
                m_vertex.set_reload_required(True)

    @property
    @overrides(AbstractCurrentSource.parameters)
    def parameters(self) -> Mapping[str, CurrentParameter]:
        return self.__parameters

    @property
    @overrides(AbstractCurrentSource.parameter_types)
    def parameter_types(self) -> Mapping[str, DataType]:
        return self.__parameter_types

    @property
    @overrides(AbstractCurrentSource.current_source_id)
    def current_source_id(self) -> int:
        return CurrentSourceIDs.NOISY_CURRENT_SOURCE.value

    @overrides(AbstractCurrentSource.get_sdram_usage_in_bytes)
    def get_sdram_usage_in_bytes(self) -> int:
        # 3 because the seed parameter has length 4
        return (len(self.__parameters) + 3) * BYTES_PER_WORD
